# Do not modify this file directly. It is auto-generated with Python.

from architectures.core import Node

class _Azure(Node):
    _provider = "azure"
    _icon_dir = "icons/azure"

class _General(Node):
    _provider = "general"
    _icon_dir = "icons/general"

class _Gcp(Node):
    _provider = "gcp"
    _icon_dir = "icons/gcp"

class _Aws(Node):
    _provider = "aws"
    _icon_dir = "icons/aws"

class _Kubernetes(Node):
    _provider = "kubernetes"
    _icon_dir = "icons/kubernetes"